# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for quantum channel representation transformations."""

import unittest
from numpy.testing import assert_allclose

from qiskit.quantum_info.states.densitymatrix import DensityMatrix
from qiskit.quantum_info.operators.operator import Operator
from qiskit.quantum_info.operators.channel.choi import Choi
from qiskit.quantum_info.operators.channel.superop import SuperOp
from qiskit.quantum_info.operators.channel.kraus import Kraus
from qiskit.quantum_info.operators.channel.stinespring import Stinespring
from qiskit.quantum_info.operators.channel.ptm import PTM
from qiskit.quantum_info.operators.channel.chi import Chi
from .channel_test_case import ChannelTestCase


class TestEvolve(ChannelTestCase):
    """Random tests for equivalence of channel evolution."""

    qubits_test_cases = (1, 2)
    repetitions = 2

    def _unitary_to_other(self, rep, qubits_test_cases, repetitions):
        """Test Operator to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim, dim)
                chan = Operator(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _other_to_operator(self, rep, qubits_test_cases, repetitions):
        """Test Other to Operator evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim, dim)
                chan = rep(Operator(mat))
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(Operator(chan)).data
                assert_allclose(rho1, rho2)

    def _choi_to_other_cp(self, rep, qubits_test_cases, repetitions):
        """Test CP Choi to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = dim * self.rand_rho(dim**2)
                chan = Choi(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _choi_to_other_noncp(self, rep, qubits_test_cases, repetitions):
        """Test CP Choi to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim**2, dim**2)
                chan = Choi(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _superop_to_other(self, rep, qubits_test_cases, repetitions):
        """Test SuperOp to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim**2, dim**2)
                chan = SuperOp(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _kraus_to_other_single(self, rep, qubits_test_cases, repetitions):
        """Test single Kraus to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                kraus = self.rand_kraus(dim, dim, dim**2)
                chan = Kraus(kraus)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _kraus_to_other_double(self, rep, qubits_test_cases, repetitions):
        """Test double Kraus to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                kraus_l = self.rand_kraus(dim, dim, dim**2)
                kraus_r = self.rand_kraus(dim, dim, dim**2)
                chan = Kraus((kraus_l, kraus_r))
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _stinespring_to_other_single(self, rep, qubits_test_cases, repetitions):
        """Test single Stinespring to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim**2, dim)
                chan = Stinespring(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _stinespring_to_other_double(self, rep, qubits_test_cases, repetitions):
        """Test double Stinespring to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat_l = self.rand_matrix(dim**2, dim)
                mat_r = self.rand_matrix(dim**2, dim)
                chan = Stinespring((mat_l, mat_r))
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _chi_to_other(self, rep, qubits_test_cases, repetitions):
        """Test Chi to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim**2, dim**2, real=True)
                chan = Chi(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def _ptm_to_other(self, rep, qubits_test_cases, repetitions):
        """Test PTM to Other evolution."""
        for nq in qubits_test_cases:
            dim = 2**nq
            for _ in range(repetitions):
                rho = self.rand_rho(dim)
                mat = self.rand_matrix(dim**2, dim**2, real=True)
                chan = PTM(mat)
                rho1 = DensityMatrix(rho).evolve(chan).data
                rho2 = DensityMatrix(rho).evolve(rep(chan)).data
                assert_allclose(rho1, rho2)

    def test_unitary_to_choi(self):
        """Test Operator to Choi evolution."""
        self._unitary_to_other(Choi, self.qubits_test_cases, self.repetitions)

    def test_unitary_to_superop(self):
        """Test Operator to SuperOp evolution."""
        self._unitary_to_other(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_unitary_to_kraus(self):
        """Test Operator to Kraus evolution."""
        self._unitary_to_other(Kraus, self.qubits_test_cases, self.repetitions)

    def test_unitary_to_stinespring(self):
        """Test Operator to Stinespring evolution."""
        self._unitary_to_other(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_unitary_to_chi(self):
        """Test Operator to Chi evolution."""
        self._unitary_to_other(Chi, self.qubits_test_cases, self.repetitions)

    def test_unitary_to_ptm(self):
        """Test Operator to PTM evolution."""
        self._unitary_to_other(PTM, self.qubits_test_cases, self.repetitions)

    def test_choi_to_operator(self):
        """Test Choi to Operator evolution."""
        self._other_to_operator(Choi, self.qubits_test_cases, self.repetitions)

    def test_choi_to_superop_cp(self):
        """Test CP Choi to SuperOp evolution."""
        self._choi_to_other_cp(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_choi_to_kraus_cp(self):
        """Test CP Choi to Kraus evolution."""
        self._choi_to_other_cp(Kraus, self.qubits_test_cases, self.repetitions)

    def test_choi_to_stinespring_cp(self):
        """Test CP Choi to Stinespring evolution."""
        self._choi_to_other_cp(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_choi_to_chi_cp(self):
        """Test CP Choi to Chi evolution."""
        self._choi_to_other_cp(Chi, self.qubits_test_cases, self.repetitions)

    def test_choi_to_ptm_cp(self):
        """Test CP Choi to PTM evolution."""
        self._choi_to_other_cp(PTM, self.qubits_test_cases, self.repetitions)

    def test_choi_to_superop_noncp(self):
        """Test CP Choi to SuperOp evolution."""
        self._choi_to_other_noncp(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_choi_to_kraus_noncp(self):
        """Test CP Choi to Kraus evolution."""
        self._choi_to_other_noncp(Kraus, self.qubits_test_cases, self.repetitions)

    def test_choi_to_stinespring_noncp(self):
        """Test CP Choi to Stinespring evolution."""
        self._choi_to_other_noncp(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_choi_to_chi_noncp(self):
        """Test Choi to Chi evolution."""
        self._choi_to_other_noncp(Chi, self.qubits_test_cases, self.repetitions)

    def test_choi_to_ptm_noncp(self):
        """Test Non-CP Choi to PTM evolution."""
        self._choi_to_other_noncp(PTM, self.qubits_test_cases, self.repetitions)

    def test_superop_to_operator(self):
        """Test SuperOp to Operator evolution."""
        self._other_to_operator(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_superop_to_choi(self):
        """Test SuperOp to Choi evolution."""
        self._superop_to_other(Choi, self.qubits_test_cases, self.repetitions)

    def test_superop_to_kraus(self):
        """Test SuperOp to Kraus evolution."""
        self._superop_to_other(Kraus, self.qubits_test_cases, self.repetitions)

    def test_superop_to_stinespring(self):
        """Test SuperOp to Stinespring evolution."""
        self._superop_to_other(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_superop_to_chi(self):
        """Test SuperOp to Chi evolution."""
        self._superop_to_other(Chi, self.qubits_test_cases, self.repetitions)

    def test_superop_to_ptm(self):
        """Test SuperOp to PTM evolution."""
        self._superop_to_other(PTM, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_operator(self):
        """Test Kraus to Operator evolution."""
        self._other_to_operator(Kraus, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_choi_single(self):
        """Test single Kraus to Choi evolution."""
        self._kraus_to_other_single(Choi, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_superop_single(self):
        """Test single Kraus to SuperOp evolution."""
        self._kraus_to_other_single(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_stinespring_single(self):
        """Test single Kraus to Stinespring evolution."""
        self._kraus_to_other_single(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_chi_single(self):
        """Test single Kraus to Chi evolution."""
        self._kraus_to_other_single(Chi, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_ptm_single(self):
        """Test single Kraus to PTM evolution."""
        self._kraus_to_other_single(PTM, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_choi_double(self):
        """Test single Kraus to Choi evolution."""
        self._kraus_to_other_double(Choi, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_superop_double(self):
        """Test single Kraus to SuperOp evolution."""
        self._kraus_to_other_double(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_stinespring_double(self):
        """Test single Kraus to Stinespring evolution."""
        self._kraus_to_other_double(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_chi_double(self):
        """Test single Kraus to Chi evolution."""
        self._kraus_to_other_double(Chi, self.qubits_test_cases, self.repetitions)

    def test_kraus_to_ptm_double(self):
        """Test single Kraus to PTM evolution."""
        self._kraus_to_other_double(PTM, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_operator(self):
        """Test Stinespring to Operator evolution."""
        self._other_to_operator(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_choi_single(self):
        """Test single Stinespring to Choi evolution."""
        self._stinespring_to_other_single(Choi, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_superop_single(self):
        """Test single Stinespring to SuperOp evolution."""
        self._stinespring_to_other_single(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_kraus_single(self):
        """Test single Stinespring to Kraus evolution."""
        self._stinespring_to_other_single(Kraus, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_chi_single(self):
        """Test single Stinespring to Chi evolution."""
        self._stinespring_to_other_single(Chi, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_ptm_single(self):
        """Test single Stinespring to PTM evolution."""
        self._stinespring_to_other_single(PTM, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_choi_double(self):
        """Test single Stinespring to Choi evolution."""
        self._stinespring_to_other_double(Choi, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_superop_double(self):
        """Test single Stinespring to SuperOp evolution."""
        self._stinespring_to_other_double(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_kraus_double(self):
        """Test single Stinespring to Kraus evolution."""
        self._stinespring_to_other_double(Kraus, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_chi_double(self):
        """Test single Stinespring to Chi evolution."""
        self._stinespring_to_other_double(Chi, self.qubits_test_cases, self.repetitions)

    def test_stinespring_to_ptm_double(self):
        """Test single Stinespring to PTM evolution."""
        self._stinespring_to_other_double(PTM, self.qubits_test_cases, self.repetitions)

    def test_ptm_to_operator(self):
        """Test PTM to Operator evolution."""
        self._other_to_operator(PTM, self.qubits_test_cases, self.repetitions)

    def test_ptm_to_choi(self):
        """Test PTM to Choi evolution."""
        self._ptm_to_other(Choi, self.qubits_test_cases, self.repetitions)

    def test_ptm_to_superop(self):
        """Test PTM to SuperOp evolution."""
        self._ptm_to_other(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_ptm_to_kraus(self):
        """Test PTM to Kraus evolution."""
        self._ptm_to_other(Kraus, self.qubits_test_cases, self.repetitions)

    def test_ptm_to_stinespring(self):
        """Test PTM to Stinespring evolution."""
        self._ptm_to_other(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_ptm_to_chi(self):
        """Test PTM to Chi evolution."""
        self._ptm_to_other(Chi, self.qubits_test_cases, self.repetitions)

    def test_chi_to_operator(self):
        """Test Chi to Operator evolution."""
        self._other_to_operator(Chi, self.qubits_test_cases, self.repetitions)

    def test_chi_to_choi(self):
        """Test Chi to Choi evolution."""
        self._chi_to_other(Choi, self.qubits_test_cases, self.repetitions)

    def test_chi_to_superop(self):
        """Test Chi to SuperOp evolution."""
        self._chi_to_other(SuperOp, self.qubits_test_cases, self.repetitions)

    def test_chi_to_kraus(self):
        """Test Chi to Kraus evolution."""
        self._chi_to_other(Kraus, self.qubits_test_cases, self.repetitions)

    def test_chi_to_stinespring(self):
        """Test Chi to Stinespring evolution."""
        self._chi_to_other(Stinespring, self.qubits_test_cases, self.repetitions)

    def test_chi_to_ptm(self):
        """Test Chi to PTM evolution."""
        self._chi_to_other(PTM, self.qubits_test_cases, self.repetitions)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Tests for the UnitarySynthesis transpiler pass.
"""

import functools
import itertools
import unittest.mock

import numpy as np
import stevedore

from qiskit.circuit import QuantumCircuit
from qiskit.converters import circuit_to_dag
from qiskit.transpiler import PassManager
from qiskit.transpiler.passes import UnitarySynthesis
from qiskit.transpiler.passes.synthesis.plugin import (
    UnitarySynthesisPlugin,
    UnitarySynthesisPluginManager,
    unitary_synthesis_plugin_names,
)
from qiskit.transpiler.passes.synthesis.unitary_synthesis import DefaultUnitarySynthesis
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class _MockExtensionManager:
    def __init__(self, plugins):
        self._plugins = {
            name: stevedore.extension.Extension(name, None, plugin, plugin())
            for name, plugin in plugins.items()
        }
        self._stevedore_manager = stevedore.ExtensionManager(
            "qiskit.unitary_synthesis", invoke_on_load=True, propagate_map_exceptions=True
        )

    def names(self):
        """Mock method to replace the stevedore names."""
        return list(self._plugins) + self._stevedore_manager.names()

    def __getitem__(self, value):
        try:
            return self._plugins[value]
        except KeyError:
            pass
        return self._stevedore_manager[value]

    def __contains__(self, value):
        return value in self._plugins or value in self._stevedore_manager

    def __iter__(self):
        return itertools.chain(self._plugins.values(), self._stevedore_manager)


class _MockPluginManager:
    def __init__(self, plugins):
        self.ext_plugins = _MockExtensionManager(plugins)


class ControllableSynthesis(UnitarySynthesisPlugin):
    """A dummy synthesis plugin, which can have its ``supports_`` properties changed to test
    different parts of the synthesis plugin interface.  By default, it accepts all keyword arguments
    and accepts all number of qubits, but if its run method is called, it just returns ``None`` to
    indicate that the gate should not be synthesized."""

    min_qubits = None
    max_qubits = None
    supported_bases = None
    supports_basis_gates = True
    supports_coupling_map = True
    supports_gate_errors = True
    supports_gate_lengths = True
    supports_natural_direction = True
    supports_pulse_optimize = True
    run = unittest.mock.MagicMock(return_value=None)

    @classmethod
    def reset(cls):
        """Reset the state of any internal mocks, and return class properties to their defaults."""
        cls.run.reset_mock()
        cls.min_qubits = None
        cls.max_qubits = None
        cls.supported_bases = None
        cls.support()

    @classmethod
    def support(cls, names=None):
        """Set the plugin to support the given keywords, and reject any that are not given.  If
        no argument is passed, then everything will be supported.  To reject everything, explicitly
        pass an empty iterable."""
        if names is None:

            def value(_name):
                return True

        else:
            names = set(names)

            def value(name):
                return name in names

        prefix = "supports_"
        for name in dir(cls):
            if name.startswith(prefix):
                setattr(cls, name, value(name[len(prefix) :]))


class TestUnitarySynthesisPlugin(QiskitTestCase):
    """Tests for the synthesis plugin interface."""

    # The proliferation of the "disable=no-members" lines are because pylint (very reasonably) can't
    # detect when we've mocked out an internal object with a MagicMock.

    MOCK_PLUGINS = {}
    DEFAULT_PLUGIN = DefaultUnitarySynthesis

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.MOCK_PLUGINS["_controllable"] = ControllableSynthesis
        # Decorate all test functions to patch the plugin manager with a fake one, which inserts our
        # testing mocks into the lookup.  It's easier to do this than to correctly register them,
        # because it's fiddly to trick importlib into registering new entry points metadata without
        # actually installing a package before the interpreter was invoked.  We don't want to leak
        # any testing classes out into public releases, so we just patch.
        decorator = unittest.mock.patch(
            "qiskit.transpiler.passes.synthesis.plugin.UnitarySynthesisPluginManager",
            functools.partial(_MockPluginManager, plugins=cls.MOCK_PLUGINS),
        )
        for name in dir(cls):
            if name.startswith("test_"):
                setattr(cls, name, decorator(getattr(cls, name)))

    def setUp(self):
        super().setUp()
        for plugin in self.MOCK_PLUGINS.values():
            plugin.reset()

    def mock_default_run_method(self):
        """Return a decorator or context manager that replaces the default synthesis plugin's run
        method with a mocked version that behaves normally, except has all the trackers attached to
        it."""
        # We need to mock out DefaultUnitarySynthesis.run, except it will actually get called as an
        # instance method, so we can't just wrap the method defined on the class, but instead we
        # need to wrap a method that has been bound to a particular instance.  This is slightly
        # fragile, because we're likely wrapping a _different_ instance, but since there are no
        # arguments to __init__, and no internal state, it should be ok.  It doesn't matter if we
        # dodged the patching of the manager class that happens elsewhere in this test suite,
        # because we're always accessing something that the patch would delegate to the inner
        # manager anyway.
        inner_default = UnitarySynthesisPluginManager().ext_plugins["default"].obj
        mock = unittest.mock.MagicMock(wraps=inner_default.run)
        return unittest.mock.patch.object(self.DEFAULT_PLUGIN, "run", mock)

    def test_mock_plugins_registered(self):
        """This is a meta test, that the internal registering mechanisms for our dummy test plugins
        exist and that we can call them."""
        registered = unitary_synthesis_plugin_names()
        for plugin in self.MOCK_PLUGINS:
            self.assertIn(plugin, registered)

    def test_call_registered_class(self):
        """Test that a non-default plugin was called."""
        qc = QuantumCircuit(2)
        qc.unitary(np.eye(4, dtype=np.complex128), [0, 1])
        pm = PassManager([UnitarySynthesis(basis_gates=["u", "cx"], method="_controllable")])
        with self.mock_default_run_method():
            pm.run(qc)
            self.DEFAULT_PLUGIN.run.assert_not_called()  # pylint: disable=no-member
        self.MOCK_PLUGINS["_controllable"].run.assert_called()

    def test_max_qubits_are_respected(self):
        """Test that the default handler gets used if the chosen plugin can't cope with a given
        unitary."""
        self.MOCK_PLUGINS["_controllable"].min_qubits = None
        self.MOCK_PLUGINS["_controllable"].max_qubits = 0
        qc = QuantumCircuit(2)
        qc.unitary(np.eye(4, dtype=np.complex128), [0, 1])
        pm = PassManager([UnitarySynthesis(basis_gates=["u", "cx"], method="_controllable")])
        with self.mock_default_run_method():
            pm.run(qc)
            self.DEFAULT_PLUGIN.run.assert_called()  # pylint: disable=no-member
        self.MOCK_PLUGINS["_controllable"].run.assert_not_called()

    def test_min_qubits_are_respected(self):
        """Test that the default handler gets used if the chosen plugin can't cope with a given
        unitary."""
        self.MOCK_PLUGINS["_controllable"].min_qubits = 3
        self.MOCK_PLUGINS["_controllable"].max_qubits = None
        qc = QuantumCircuit(2)
        qc.unitary(np.eye(4, dtype=np.complex128), [0, 1])
        pm = PassManager([UnitarySynthesis(basis_gates=["u", "cx"], method="_controllable")])
        with self.mock_default_run_method():
            pm.run(qc)
            self.DEFAULT_PLUGIN.run.assert_called()  # pylint: disable=no-member
        self.MOCK_PLUGINS["_controllable"].run.assert_not_called()

    def test_all_keywords_passed_to_default_on_fallback(self):
        """Test that all the keywords that the default synthesis plugin needs are passed to it, even
        if the chosen method doesn't support them."""
        # Set the mock plugin to reject all keyword arguments, but also be unable to handle
        # operators of any numbers of qubits.  This will cause fallback to the default handler,
        # which should receive a full set of keywords, still.
        self.MOCK_PLUGINS["_controllable"].min_qubits = np.inf
        self.MOCK_PLUGINS["_controllable"].max_qubits = 0
        self.MOCK_PLUGINS["_controllable"].support([])
        qc = QuantumCircuit(2)
        qc.unitary(np.eye(4, dtype=np.complex128), [0, 1])
        pm = PassManager([UnitarySynthesis(basis_gates=["u", "cx"], method="_controllable")])
        with self.mock_default_run_method():
            pm.run(qc)
            self.DEFAULT_PLUGIN.run.assert_called()  # pylint: disable=no-member
            # This access should be `run.call_args.kwargs`, but the namedtuple access wasn't added
            # until Python 3.8.
            call_kwargs = self.DEFAULT_PLUGIN.run.call_args[1]  # pylint: disable=no-member
        expected_kwargs = [
            "basis_gates",
            "coupling_map",
            "gate_errors_by_qubit",
            "gate_lengths_by_qubit",
            "natural_direction",
            "pulse_optimize",
        ]
        for kwarg in expected_kwargs:
            self.assertIn(kwarg, call_kwargs)
        self.MOCK_PLUGINS["_controllable"].run.assert_not_called()

    def test_config_passed_to_non_default(self):
        """Test that a specified non-default plugin gets a config dict passed to it."""
        self.MOCK_PLUGINS["_controllable"].min_qubits = 0
        self.MOCK_PLUGINS["_controllable"].max_qubits = np.inf
        self.MOCK_PLUGINS["_controllable"].support([])
        qc = QuantumCircuit(2)
        qc.unitary(np.eye(4, dtype=np.complex128), [0, 1])
        return_dag = circuit_to_dag(qc)
        plugin_config = {"option_a": 3.14, "option_b": False}
        pm = PassManager(
            [
                UnitarySynthesis(
                    basis_gates=["u", "cx"], method="_controllable", plugin_config=plugin_config
                )
            ]
        )
        with unittest.mock.patch.object(
            ControllableSynthesis, "run", return_value=return_dag
        ) as plugin_mock:
            pm.run(qc)
            plugin_mock.assert_called()
            # This access should be `run.call_args.kwargs`, but the namedtuple access wasn't added
            # until Python 3.8.
            call_kwargs = plugin_mock.call_args[1]
        expected_kwargs = [
            "config",
        ]
        for kwarg in expected_kwargs:
            self.assertIn(kwarg, call_kwargs)
        self.assertEqual(call_kwargs["config"], plugin_config)

    def test_config_not_passed_to_default_on_fallback(self):
        """Test that all the keywords that the default synthesis plugin needs are passed to it,
        and if if config is specified it is not passed to the default."""
        # Set the mock plugin to reject all keyword arguments, but also be unable to handle
        # operators of any numbers of qubits.  This will cause fallback to the default handler,
        # which should receive a full set of keywords, still.
        self.MOCK_PLUGINS["_controllable"].min_qubits = np.inf
        self.MOCK_PLUGINS["_controllable"].max_qubits = 0
        self.MOCK_PLUGINS["_controllable"].support([])
        qc = QuantumCircuit(2)
        qc.unitary(np.eye(4, dtype=np.complex128), [0, 1])
        plugin_config = {"option_a": 3.14, "option_b": False}
        pm = PassManager(
            [
                UnitarySynthesis(
                    basis_gates=["u", "cx"], method="_controllable", plugin_config=plugin_config
                )
            ]
        )
        with self.mock_default_run_method():
            pm.run(qc)
            self.DEFAULT_PLUGIN.run.assert_called()  # pylint: disable=no-member
            # This access should be `run.call_args.kwargs`, but the namedtuple access wasn't added
            # until Python 3.8.
            call_kwargs = self.DEFAULT_PLUGIN.run.call_args[1]  # pylint: disable=no-member
        expected_kwargs = [
            "basis_gates",
            "coupling_map",
            "gate_errors_by_qubit",
            "gate_lengths_by_qubit",
            "natural_direction",
            "pulse_optimize",
        ]
        for kwarg in expected_kwargs:
            self.assertIn(kwarg, call_kwargs)
        self.MOCK_PLUGINS["_controllable"].run.assert_not_called()
        self.assertNotIn("config", call_kwargs)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module containing multi-controlled circuits synthesis"""

from .mcmt_vchain import synth_mcmt_vchain
from .mcx_synthesis import (
    synth_mcx_n_dirty_i15,
    synth_mcx_n_clean_m15,
    synth_mcx_1_clean_b95,
    synth_mcx_gray_code,
    synth_mcx_noaux_v24,
    synth_c3x,
    synth_c4x,
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Simple circuit constructors for Weyl reflections.
"""

from __future__ import annotations
import numpy as np

from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.library.standard_gates import RXGate, RYGate, RZGate


reflection_options = {
    "no reflection": ([1, 1, 1], 1, []),
    "reflect XX, YY": ([-1, -1, 1], 1, [RZGate]),
    "reflect XX, ZZ": ([-1, 1, -1], 1, [RYGate]),
    "reflect YY, ZZ": ([1, -1, -1], 1, [RXGate]),
}
"""
A table of available reflection transformations on canonical coordinates.
Entries take the form

    readable_name: (reflection scalars, global phase, [gate constructors]),

where reflection scalars (a, b, c) model the map (x, y, z) |-> (ax, by, cz),
global phase is a complex unit, and gate constructors are applied in sequence
and by conjugation to the first qubit and are passed pi as a parameter.
"""

shift_options = {
    "no shift": ([0, 0, 0], 1, []),
    "Z shift": ([0, 0, 1], 1j, [RZGate]),
    "Y shift": ([0, 1, 0], -1j, [RYGate]),
    "Y,Z shift": ([0, 1, 1], 1, [RYGate, RZGate]),
    "X shift": ([1, 0, 0], -1j, [RXGate]),
    "X,Z shift": ([1, 0, 1], 1, [RXGate, RZGate]),
    "X,Y shift": ([1, 1, 0], -1, [RXGate, RYGate]),
    "X,Y,Z shift": ([1, 1, 1], -1j, [RXGate, RYGate, RZGate]),
}
"""
A table of available shift transformations on canonical coordinates.  Entries
take the form

    readable name: (shift scalars, global phase, [gate constructors]),

where shift scalars model the map

    (x, y, z) |-> (x + a pi / 2, y + b pi / 2, z + c pi / 2) ,

global phase is a complex unit, and gate constructors are applied to the first
and second qubits and are passed pi as a parameter.
"""


def apply_reflection(reflection_name, coordinate):
    """
    Given a reflection type and a canonical coordinate, applies the reflection
    and describes a circuit which enacts the reflection + a global phase shift.
    """
    reflection_scalars, reflection_phase_shift, source_reflection_gates = reflection_options[
        reflection_name
    ]
    reflected_coord = [x * y for x, y in zip(reflection_scalars, coordinate)]
    source_reflection = QuantumCircuit(2)
    for gate in source_reflection_gates:
        source_reflection.append(gate(np.pi), [0])

    return reflected_coord, source_reflection, reflection_phase_shift


def apply_shift(shift_name, coordinate):
    """
    Given a shift type and a canonical coordinate, applies the shift and
    describes a circuit which enacts the shift + a global phase shift.
    """
    shift_scalars, shift_phase_shift, source_shift_gates = shift_options[shift_name]
    shifted_coord = [np.pi / 2 * x + y for x, y in zip(shift_scalars, coordinate)]

    source_shift = QuantumCircuit(2)
    for gate in source_shift_gates:
        source_shift.append(gate(np.pi), [0])
        source_shift.append(gate(np.pi), [1])

    return shifted_coord, source_shift, shift_phase_shift


def canonical_rotation_circuit(first_index, second_index):
    """
    Given a pair of distinct indices 0 ≤ (first_index, second_index) ≤ 2,
    produces a two-qubit circuit which rotates a canonical gate

        a0 XX + a1 YY + a2 ZZ

    into

        a[first] XX + a[second] YY + a[other] ZZ .
    """
    conj = QuantumCircuit(2)

    if (0, 1) == (first_index, second_index):
        pass  # no need to do anything
    elif (0, 2) == (first_index, second_index):
        conj.rx(-np.pi / 2, [0])
        conj.rx(np.pi / 2, [1])
    elif (1, 0) == (first_index, second_index):
        conj.rz(-np.pi / 2, [0])
        conj.rz(-np.pi / 2, [1])
    elif (1, 2) == (first_index, second_index):
        conj.rz(np.pi / 2, [0])
        conj.rz(np.pi / 2, [1])
        conj.ry(np.pi / 2, [0])
        conj.ry(-np.pi / 2, [1])
    elif (2, 0) == (first_index, second_index):
        conj.rz(np.pi / 2, [0])
        conj.rz(np.pi / 2, [1])
        conj.rx(np.pi / 2, [0])
        conj.rx(-np.pi / 2, [1])
    elif (2, 1) == (first_index, second_index):
        conj.ry(np.pi / 2, [0])
        conj.ry(-np.pi / 2, [1])

    return conj



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test OptimizeSwapBeforeMeasure pass"""

import unittest

from qiskit import QuantumRegister, QuantumCircuit, ClassicalRegister
from qiskit.passmanager.flow_controllers import DoWhileController
from qiskit.transpiler import PassManager
from qiskit.transpiler.passes import OptimizeSwapBeforeMeasure, DAGFixedPoint
from qiskit.converters import circuit_to_dag
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestOptimizeSwapBeforeMeasure(QiskitTestCase):
    """Test swap-followed-by-measure optimizations."""

    def test_optimize_1swap_1measure(self):
        """Remove a single swap
        qr0:--X--m--       qr0:----
              |  |
        qr1:--X--|--  ==>  qr1:--m-
                 |               |
        cr0:-----.--       cr0:--.-
        """
        qr = QuantumRegister(2, "qr")
        cr = ClassicalRegister(1, "cr")
        circuit = QuantumCircuit(qr, cr)
        circuit.swap(qr[0], qr[1])
        circuit.measure(qr[0], cr[0])
        dag = circuit_to_dag(circuit)

        expected = QuantumCircuit(qr, cr)
        expected.measure(qr[1], cr[0])

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(expected), after)

    def test_optimize_1swap_2measure(self):
        """Remove a single swap affecting two measurements
        qr0:--X--m--         qr0:--m----
              |  |                 |
        qr1:--X--|--m   ==>  qr1:--|--m-
                 |  |              |  |
        cr0:-----.--|--      cr0:--|--.-
        cr1:--------.--      cr1:--.----
        """
        qr = QuantumRegister(2, "qr")
        cr = ClassicalRegister(2, "cr")
        circuit = QuantumCircuit(qr, cr)
        circuit.swap(qr[0], qr[1])
        circuit.measure(qr[0], cr[0])
        circuit.measure(qr[1], cr[1])
        dag = circuit_to_dag(circuit)

        expected = QuantumCircuit(qr, cr)
        expected.measure(qr[1], cr[0])
        expected.measure(qr[0], cr[1])

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(expected), after)

    def test_optimize_3swap_3measure(self):
        """Remove three swaps affecting three measurements
                ┌─┐         ┌─┐             ┌─┐
        q_0: ─X─┤M├────X──X─┤M├     q_0: ───┤M├───
              │ └╥┘┌─┐ │  │ └╥┘          ┌─┐└╥┘┌─┐
        q_1: ─X──╫─┤M├─┼──┼──╫─     q_1: ┤M├─╫─┤M├
                 ║ └╥┘ │  │  ║  ==>      └╥┘ ║ └╥┘
        q_2: ────╫──╫──X──X──╫─     q_2: ─╫──╫──╫─
                 ║  ║        ║            ║  ║  ║
        c: 2/════╩══╩════════╩═     c: 2/═╩══╩══╩═
                 0  1        0            0  1  0
        """

        qc = QuantumCircuit(3, 2)
        qc.swap(0, 1)
        qc.measure(0, 0)
        qc.swap(0, 2)
        qc.measure(1, 1)
        qc.swap(0, 2)
        qc.measure(0, 0)
        dag = circuit_to_dag(qc)

        expected = QuantumCircuit(3, 2)
        expected.measure(1, 0)
        expected.measure(0, 1)
        expected.measure(1, 0)

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(expected), after)

    def test_optimize_nswap_nmeasure(self):
        """Remove severals swap affecting multiple measurements
                            ┌─┐                                                   ┌─┐
        q_0: ─X──X─────X────┤M├─────────────────────────────────       q_0: ──────┤M├───────────────
              │  │     │    └╥┘         ┌─┐                                    ┌─┐└╥┘
        q_1: ─X──X──X──X──X──╫─────X────┤M├─────────────────────       q_1: ───┤M├─╫────────────────
                    │     │  ║     │    └╥┘      ┌─┐                        ┌─┐└╥┘ ║
        q_2: ───────X──X──X──╫──X──X─────╫──X────┤M├────────────       q_2: ┤M├─╫──╫────────────────
                       │     ║  │        ║  │    └╥┘┌─┐                     └╥┘ ║  ║    ┌─┐
        q_3: ─X─────X──X─────╫──X──X──X──╫──X─────╫─┤M├─────────       q_3: ─╫──╫──╫────┤M├─────────
              │     │        ║     │  │  ║        ║ └╥┘┌─┐                   ║  ║  ║    └╥┘      ┌─┐
        q_4: ─X──X──X──X─────╫──X──X──X──╫──X─────╫──╫─┤M├──────  ==>  q_4: ─╫──╫──╫─────╫───────┤M├
                 │     │     ║  │        ║  │     ║  ║ └╥┘┌─┐                ║  ║  ║ ┌─┐ ║       └╥┘
        q_5: ────X──X──X──X──╫──X──X─────╫──X──X──╫──╫──╫─┤M├───       q_5: ─╫──╫──╫─┤M├─╫────────╫─
                    │     │  ║     │     ║     │  ║  ║  ║ └╥┘┌─┐             ║  ║  ║ └╥┘ ║ ┌─┐    ║
        q_6: ─X──X──X──X──X──╫──X──X─────╫─────X──╫──╫──╫──╫─┤M├       q_6: ─╫──╫──╫──╫──╫─┤M├────╫─
              │  │     │     ║  │ ┌─┐    ║        ║  ║  ║  ║ └╥┘             ║  ║  ║  ║  ║ └╥┘┌─┐ ║
        q_7: ─X──X─────X─────╫──X─┤M├────╫────────╫──╫──╫──╫──╫─       q_7: ─╫──╫──╫──╫──╫──╫─┤M├─╫─
                             ║    └╥┘    ║        ║  ║  ║  ║  ║              ║  ║  ║  ║  ║  ║ └╥┘ ║
        c: 8/════════════════╩═════╩═════╩════════╩══╩══╩══╩══╩═       c: 8/═╩══╩══╩══╩══╩══╩══╩══╩═
                             0     7     1        2  3  4  5  6              0  1  2  3  4  5  6  7
        """
        circuit = QuantumCircuit(8, 8)
        circuit.swap(3, 4)
        circuit.swap(6, 7)
        circuit.swap(0, 1)
        circuit.swap(6, 7)
        circuit.swap(4, 5)
        circuit.swap(0, 1)
        circuit.swap(5, 6)
        circuit.swap(3, 4)
        circuit.swap(1, 2)
        circuit.swap(6, 7)
        circuit.swap(4, 5)
        circuit.swap(2, 3)
        circuit.swap(0, 1)
        circuit.swap(5, 6)
        circuit.swap(1, 2)
        circuit.swap(6, 7)
        circuit.swap(4, 5)
        circuit.swap(2, 3)
        circuit.swap(3, 4)
        circuit.swap(3, 4)
        circuit.swap(5, 6)
        circuit.swap(1, 2)
        circuit.swap(4, 5)
        circuit.swap(2, 3)
        circuit.swap(5, 6)
        circuit.measure(range(8), range(8))
        dag = circuit_to_dag(circuit)

        expected = QuantumCircuit(8, 8)
        expected.measure(0, 2)
        expected.measure(1, 1)
        expected.measure(2, 0)
        expected.measure(3, 4)
        expected.measure(4, 7)
        expected.measure(5, 3)
        expected.measure(6, 5)
        expected.measure(7, 6)

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(expected), after)

    def test_optimize_nswap_nmeasure_2(self):
        """Remove multiple swaps affecting multiple measurements
                   ┌─┐┌─┐                             ┌─┐┌─┐┌─┐┌─┐┌─┐
        q_0: ─X──X─┤M├┤M├─X──────────────X─      q_0: ┤M├┤M├┤M├┤M├┤M├
              │  │ └╥┘└╥┘ │       ┌─┐    │            └╥┘└╥┘└╥┘└╥┘└╥┘
        q_1: ─X──X──╫──╫──┼─────X─┤M├─X──X─      q_1: ─╫──╫──╫──╫──╫─
                    ║  ║  │ ┌─┐ │ └╥┘ │ ┌─┐  ==>       ║  ║  ║  ║  ║
        q_2: ───────╫──╫──X─┤M├─X──╫──X─┤M├      q_2: ─╫──╫──╫──╫──╫─
                    ║  ║    └╥┘    ║    └╥┘            ║  ║  ║  ║  ║
        c: 1/═══════╩══╩═════╩═════╩═════╩═      c: 1/═╩══╩══╩══╩══╩═
                    0  0     0     0     0             0  0  0  0  0
        """

        qc = QuantumCircuit(3, 1)
        qc.swap(0, 1)
        qc.swap(0, 1)
        qc.measure(0, 0)
        qc.measure(0, 0)
        qc.swap(2, 0)
        qc.measure(2, 0)
        qc.swap(1, 2)
        qc.measure(1, 0)
        qc.swap(1, 2)
        qc.swap(0, 1)
        qc.measure(2, 0)
        dag = circuit_to_dag(qc)

        expected = QuantumCircuit(3, 1)
        expected.measure(0, 0)
        expected.measure(0, 0)
        expected.measure(0, 0)
        expected.measure(0, 0)
        expected.measure(0, 0)

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(expected), after)

    def test_cannot_optimize(self):
        """Cannot optimize when swap is not at the end in all of the successors
        qr0:--X-----m--
              |     |
        qr1:--X-[H]-|--
                    |
        cr0:--------.--
        """
        qr = QuantumRegister(2, "qr")
        cr = ClassicalRegister(1, "cr")
        circuit = QuantumCircuit(qr, cr)
        circuit.swap(qr[0], qr[1])
        circuit.h(qr[1])
        circuit.measure(qr[0], cr[0])
        dag = circuit_to_dag(circuit)

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(circuit), after)

    def test_if_else(self):
        """Test that the pass recurses into a simple if-else."""
        pass_ = OptimizeSwapBeforeMeasure()

        base_test = QuantumCircuit(2, 1)
        base_test.swap(0, 1)
        base_test.measure(0, 0)

        base_expected = QuantumCircuit(2, 1)
        base_expected.measure(1, 0)

        test = QuantumCircuit(2, 1)
        test.if_else(
            (test.clbits[0], True), base_test.copy(), base_test.copy(), test.qubits, test.clbits
        )

        expected = QuantumCircuit(2, 1)
        expected.if_else(
            (expected.clbits[0], True),
            base_expected.copy(),
            base_expected.copy(),
            expected.qubits,
            expected.clbits,
        )

        self.assertEqual(pass_(test), expected)

    def test_nested_control_flow(self):
        """Test that the pass recurses into nested control flow."""
        pass_ = OptimizeSwapBeforeMeasure()

        base_test = QuantumCircuit(2, 1)
        base_test.swap(0, 1)
        base_test.measure(0, 0)

        base_expected = QuantumCircuit(2, 1)
        base_expected.measure(1, 0)

        body_test = QuantumCircuit(2, 1)
        body_test.for_loop((0,), None, base_expected.copy(), body_test.qubits, body_test.clbits)

        body_expected = QuantumCircuit(2, 1)
        body_expected.for_loop(
            (0,), None, base_expected.copy(), body_expected.qubits, body_expected.clbits
        )

        test = QuantumCircuit(2, 1)
        test.while_loop((test.clbits[0], True), body_test, test.qubits, test.clbits)

        expected = QuantumCircuit(2, 1)
        expected.while_loop(
            (expected.clbits[0], True), body_expected, expected.qubits, expected.clbits
        )

        self.assertEqual(pass_(test), expected)


class TestOptimizeSwapBeforeMeasureFixedPoint(QiskitTestCase):
    """Test swap-followed-by-measure optimizations in a transpiler, using fixed point."""

    def test_optimize_undone_swap(self):
        """Remove redundant swap
        qr0:--X--X--m--       qr0:--m---
              |  |  |               |
        qr1:--X--X--|--  ==>  qr1:--|--
                    |               |
        cr0:--------.--       cr0:--.--
        """
        qr = QuantumRegister(2, "qr")
        cr = ClassicalRegister(1, "cr")
        circuit = QuantumCircuit(qr, cr)
        circuit.swap(qr[0], qr[1])
        circuit.swap(qr[0], qr[1])
        circuit.measure(qr[0], cr[0])

        expected = QuantumCircuit(qr, cr)
        expected.measure(qr[0], cr[0])

        pass_manager = PassManager()
        pass_manager.append(
            DoWhileController(
                [OptimizeSwapBeforeMeasure(), DAGFixedPoint()],
                do_while=lambda property_set: not property_set["dag_fixed_point"],
            )
        )
        after = pass_manager.run(circuit)

        self.assertEqual(expected, after)

    def test_optimize_overlap_swap(self):
        """Remove two swaps that overlap
        qr0:--X--------       qr0:--m--
              |                     |
        qr1:--X--X-----       qr1:--|--
                 |       ==>        |
        qr2:-----X--m--       qr2:--|--
                    |               |
        cr0:--------.--       cr0:--.--
        """
        qr = QuantumRegister(3, "qr")
        cr = ClassicalRegister(1, "cr")
        circuit = QuantumCircuit(qr, cr)
        circuit.swap(qr[0], qr[1])
        circuit.swap(qr[1], qr[2])
        circuit.measure(qr[2], cr[0])

        expected = QuantumCircuit(qr, cr)
        expected.measure(qr[0], cr[0])

        pass_manager = PassManager()
        pass_manager.append(
            DoWhileController(
                [OptimizeSwapBeforeMeasure(), DAGFixedPoint()],
                do_while=lambda property_set: not property_set["dag_fixed_point"],
            )
        )
        after = pass_manager.run(circuit)

        self.assertEqual(expected, after)

    def test_no_optimize_swap_with_condition(self):
        """Do not remove swap if it has a condition
        qr0:--X--m--       qr0:--X--m--
              |  |               |  |
        qr1:--X--|--  ==>  qr1:--X--|--
              |  |               |  |
        cr0:--1--.--       cr0:--1--.--
        """
        qr = QuantumRegister(2, "qr")
        cr = ClassicalRegister(1, "cr")
        circuit = QuantumCircuit(qr, cr)
        with self.assertWarns(DeprecationWarning):
            circuit.swap(qr[0], qr[1]).c_if(cr, 1)
        circuit.measure(qr[0], cr[0])
        dag = circuit_to_dag(circuit)

        expected = QuantumCircuit(qr, cr)
        with self.assertWarns(DeprecationWarning):
            expected.swap(qr[0], qr[1]).c_if(cr, 1)
        expected.measure(qr[0], cr[0])

        pass_ = OptimizeSwapBeforeMeasure()
        after = pass_.run(dag)

        self.assertEqual(circuit_to_dag(expected), after)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
A library of known embodiments of RXXGate in terms of other gates,
for some generic or specific angles.

TODO: discover these automatically from the gates' algebraic definition
"""

from __future__ import annotations
import numpy as np

from qiskit.circuit import QuantumCircuit, Parameter
from qiskit.circuit.library.standard_gates import (
    RXXGate,
    RZZGate,
    RZXGate,
    RYYGate,
    CRZGate,
    CRXGate,
    CRYGate,
    CPhaseGate,
    CZGate,
    CXGate,
    CYGate,
    CHGate,
    ECRGate,
)


rxx_circuit = QuantumCircuit(2)
theta = Parameter("θ")
rxx_circuit.rxx(theta, 0, 1)

rzz_circuit = QuantumCircuit(2)
theta = Parameter("θ")
rzz_circuit.h(0)
rzz_circuit.h(1)
rzz_circuit.rzz(theta, 0, 1)
rzz_circuit.h(0)
rzz_circuit.h(1)

rzx_circuit = QuantumCircuit(2)
rzx_circuit.h(0)
rzx_circuit.rzx(theta, 0, 1)
rzx_circuit.h(0)

ryy_circuit = QuantumCircuit(2)
ryy_circuit.s(0)
ryy_circuit.s(1)
ryy_circuit.ryy(theta, 0, 1)
ryy_circuit.sdg(0)
ryy_circuit.sdg(1)

cphase_circuit = QuantumCircuit(2)
cphase_circuit.h(0)
cphase_circuit.h(1)
cphase_circuit.cp(-2 * theta, 0, 1)
cphase_circuit.rz(theta, 0)
cphase_circuit.rz(theta, 1)
cphase_circuit.h(0)
cphase_circuit.h(1)
cphase_circuit.global_phase += theta / 2

crz_circuit = QuantumCircuit(2)
crz_circuit.h(0)
crz_circuit.h(1)
crz_circuit.crz(-2 * theta, 0, 1)
crz_circuit.rz(theta, 1)
crz_circuit.h(0)
crz_circuit.h(1)

crx_circuit = QuantumCircuit(2)
crx_circuit.h(0)
crx_circuit.crx(-2 * theta, 0, 1)
crx_circuit.rx(theta, 1)
crx_circuit.h(0)

cry_circuit = QuantumCircuit(2)
cry_circuit.h(0)
cry_circuit.s(1)
cry_circuit.cry(-2 * theta, 0, 1)
cry_circuit.ry(theta, 1)
cry_circuit.h(0)
cry_circuit.sdg(1)

cz_circuit = QuantumCircuit(2)
cz_circuit.h(0)
cz_circuit.h(1)
cz_circuit.cz(0, 1)
cz_circuit.s(0)
cz_circuit.s(1)
cz_circuit.h(0)
cz_circuit.h(1)
cz_circuit.global_phase -= np.pi / 4

cx_circuit = QuantumCircuit(2)
cx_circuit.h(0)
cx_circuit.cx(0, 1)
cx_circuit.s(0)
cx_circuit.sx(1)
cx_circuit.h(0)
cx_circuit.global_phase -= np.pi / 4

cy_circuit = QuantumCircuit(2)
cy_circuit.h(0)
cy_circuit.s(1)
cy_circuit.cy(0, 1)
cy_circuit.s(0)
cy_circuit.sdg(1)
cy_circuit.sx(1)
cy_circuit.h(0)
cy_circuit.global_phase -= np.pi / 4

ch_circuit = QuantumCircuit(2)
ch_circuit.h(0)
ch_circuit.tdg(1)
ch_circuit.h(1)
ch_circuit.sdg(1)
ch_circuit.ch(0, 1)
ch_circuit.s(0)
ch_circuit.s(1)
ch_circuit.h(1)
ch_circuit.t(1)
ch_circuit.sx(1)
ch_circuit.h(0)
ch_circuit.global_phase -= np.pi / 4

ecr_circuit = QuantumCircuit(2)
ecr_circuit.h(0)
ecr_circuit.s(0)
ecr_circuit.x(0)
ecr_circuit.x(1)
ecr_circuit.ecr(0, 1)
ecr_circuit.s(0)
ecr_circuit.h(0)
ecr_circuit.global_phase -= np.pi / 2

XXEmbodiments = {
    RXXGate: rxx_circuit,
    RYYGate: ryy_circuit,
    RZZGate: rzz_circuit,
    RZXGate: rzx_circuit,
    CRXGate: crx_circuit,
    CRYGate: cry_circuit,
    CRZGate: crz_circuit,
    CPhaseGate: cphase_circuit,
    CXGate: cx_circuit,
    CYGate: cy_circuit,
    CZGate: cz_circuit,
    CHGate: ch_circuit,
    ECRGate: ecr_circuit,
}



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Helper function for converting a dag circuit to a dag dependency"""
from qiskit.dagcircuit.dagdependency_v2 import _DAGDependencyV2


def _dag_to_dagdependency_v2(dag):
    """Build a ``_DAGDependencyV2`` object from a ``DAGCircuit``.

    Args:
        dag (DAGCircuit): the input dag.

    Return:
        _DAGDependencyV2: the DAG representing the input circuit as a dag dependency.
    """
    dagdependency = _DAGDependencyV2()
    dagdependency.name = dag.name
    dagdependency.metadata = dag.metadata
    dagdependency.global_phase = dag.global_phase
    dagdependency.calibrations = dag._calibrations_prop

    dagdependency.add_qubits(dag.qubits)
    dagdependency.add_clbits(dag.clbits)

    for register in dag.qregs.values():
        dagdependency.add_qreg(register)

    for register in dag.cregs.values():
        dagdependency.add_creg(register)

    for node in dag.topological_op_nodes():
        dagdependency.apply_operation_back(node.op.copy(), node.qargs, node.cargs)

    return dagdependency



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the Solovay Kitaev transpilation pass."""

import os
import unittest
import math
import tempfile
import numpy as np
import scipy

from ddt import ddt, data

from qiskit import transpile
from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import TGate, TdgGate, HGate, SGate, SdgGate, IGate, QFT
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.quantum_info import Operator
from qiskit.synthesis.discrete_basis.generate_basis_approximations import (
    generate_basic_approximations,
)
from qiskit.synthesis.discrete_basis.commutator_decompose import commutator_decompose
from qiskit.synthesis.discrete_basis.gate_sequence import GateSequence
from qiskit.transpiler import PassManager
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.passes import UnitarySynthesis, Collect1qRuns, ConsolidateBlocks
from qiskit.transpiler.passes.synthesis import SolovayKitaev, SolovayKitaevSynthesis
from test import QiskitTestCase  # pylint: disable=wrong-import-order


def _trace_distance(circuit1, circuit2):
    """Return the trace distance of the two input circuits."""
    op1, op2 = Operator(circuit1), Operator(circuit2)
    return 0.5 * np.trace(scipy.linalg.sqrtm(np.conj(op1 - op2).T.dot(op1 - op2))).real


def _generate_x_rotation(angle: float) -> np.ndarray:
    return np.array(
        [[1, 0, 0], [0, math.cos(angle), -math.sin(angle)], [0, math.sin(angle), math.cos(angle)]]
    )


def _generate_y_rotation(angle: float) -> np.ndarray:
    return np.array(
        [[math.cos(angle), 0, math.sin(angle)], [0, 1, 0], [-math.sin(angle), 0, math.cos(angle)]]
    )


def _generate_z_rotation(angle: float) -> np.ndarray:
    return np.array(
        [[math.cos(angle), -math.sin(angle), 0], [math.sin(angle), math.cos(angle), 0], [0, 0, 1]]
    )


def is_so3_matrix(array: np.ndarray) -> bool:
    """Check if the input array is a SO(3) matrix."""
    if array.shape != (3, 3):
        return False

    if abs(np.linalg.det(array) - 1.0) > 1e-10:
        return False

    if False in np.isreal(array):
        return False

    return True


@ddt
class TestSolovayKitaev(QiskitTestCase):
    """Test the Solovay Kitaev algorithm and transformation pass."""

    def setUp(self):
        super().setUp()

        self.basic_approx = generate_basic_approximations([HGate(), TGate(), TdgGate()], 3)

    def test_unitary_synthesis(self):
        """Test the unitary synthesis transpiler pass with Solovay-Kitaev."""
        circuit = QuantumCircuit(2)
        circuit.rx(0.8, 0)
        circuit.cx(0, 1)
        circuit.x(1)

        _1q = Collect1qRuns()
        _cons = ConsolidateBlocks()
        _synth = UnitarySynthesis(["h", "s"], method="sk")
        passes = PassManager([_1q, _cons, _synth])
        compiled = passes.run(circuit)

        diff = np.linalg.norm(Operator(compiled) - Operator(circuit))
        self.assertLess(diff, 1)
        self.assertEqual(set(compiled.count_ops().keys()), {"h", "s", "cx"})

    def test_plugin(self):
        """Test calling the plugin directly."""
        circuit = QuantumCircuit(1)
        circuit.rx(0.8, 0)

        unitary = Operator(circuit).data

        plugin = SolovayKitaevSynthesis()
        out = plugin.run(unitary, basis_gates=["h", "s"])

        reference = QuantumCircuit(1, global_phase=3 * np.pi / 4)
        reference.h(0)
        reference.s(0)
        reference.h(0)

        self.assertEqual(dag_to_circuit(out), reference)

    def test_generating_default_approximation(self):
        """Test the approximation set is generated by default."""
        skd = SolovayKitaev()
        circuit = QuantumCircuit(1)
        dummy = skd(circuit)

        self.assertIsNotNone(skd._sk.basic_approximations)

    def test_i_returns_empty_circuit(self):
        """Test that ``SolovayKitaev`` returns an empty circuit when
        it approximates the I-gate."""
        circuit = QuantumCircuit(1)
        circuit.id(0)

        skd = SolovayKitaev(3, self.basic_approx)

        decomposed_circuit = skd(circuit)
        self.assertEqual(QuantumCircuit(1), decomposed_circuit)

    def test_exact_decomposition_acts_trivially(self):
        """Test that the a circuit that can be represented exactly is represented exactly."""
        circuit = QuantumCircuit(1)
        circuit.t(0)
        circuit.h(0)
        circuit.tdg(0)

        synth = SolovayKitaev(3, self.basic_approx)

        dag = circuit_to_dag(circuit)
        decomposed_dag = synth.run(dag)
        decomposed_circuit = dag_to_circuit(decomposed_dag)
        self.assertEqual(circuit, decomposed_circuit)

    def test_fails_with_no_to_matrix(self):
        """Test failer if gate does not have to_matrix."""
        circuit = QuantumCircuit(1)
        circuit.initialize("0")

        synth = SolovayKitaev(3, self.basic_approx)

        dag = circuit_to_dag(circuit)

        with self.assertRaises(TranspilerError) as cm:
            _ = synth.run(dag)

        self.assertEqual(
            "SolovayKitaev does not support gate without to_matrix method: initialize",
            cm.exception.message,
        )

    def test_str_basis_gates(self):
        """Test specifying the basis gates by string works."""
        circuit = QuantumCircuit(1)
        circuit.rx(0.8, 0)

        basic_approx = generate_basic_approximations(["h", "t", "s"], 3)
        synth = SolovayKitaev(2, basic_approx)

        dag = circuit_to_dag(circuit)
        discretized = dag_to_circuit(synth.run(dag))

        reference = QuantumCircuit(1, global_phase=7 * np.pi / 8)
        reference.h(0)
        reference.t(0)
        reference.h(0)

        self.assertEqual(discretized, reference)

    def test_approximation_on_qft(self):
        """Test the Solovay-Kitaev decomposition on the QFT circuit."""
        qft = QFT(3)
        transpiled = transpile(qft, basis_gates=["u", "cx"], optimization_level=1)

        skd = SolovayKitaev(1)

        with self.subTest("1 recursion"):
            discretized = skd(transpiled)
            self.assertLess(_trace_distance(transpiled, discretized), 15)

        skd.recursion_degree = 2
        with self.subTest("2 recursions"):
            discretized = skd(transpiled)
            self.assertLess(_trace_distance(transpiled, discretized), 7)

    def test_u_gates_work(self):
        """Test SK works on Qiskit's UGate.

        Regression test of Qiskit/qiskit-terra#9437.
        """
        circuit = QuantumCircuit(1)
        circuit.u(np.pi / 2, -np.pi, -np.pi, 0)
        circuit.u(np.pi / 2, np.pi / 2, -np.pi, 0)
        circuit.u(-np.pi / 4, 0, -np.pi / 2, 0)
        circuit.u(np.pi / 4, -np.pi / 16, 0, 0)
        circuit.u(0, 0, np.pi / 16, 0)
        circuit.u(0, np.pi / 4, np.pi / 4, 0)
        circuit.u(np.pi / 2, 0, -15 * np.pi / 16, 0)
        circuit.p(-np.pi / 4, 0)
        circuit.p(np.pi / 4, 0)
        circuit.u(np.pi / 2, 0, -3 * np.pi / 4, 0)
        circuit.u(0, 0, -np.pi / 16, 0)
        circuit.u(np.pi / 2, 0, 15 * np.pi / 16, 0)

        depth = 4
        basis_gates = ["h", "t", "tdg", "s", "z"]
        gate_approx_library = generate_basic_approximations(basis_gates=basis_gates, depth=depth)

        skd = SolovayKitaev(recursion_degree=2, basic_approximations=gate_approx_library)
        discretized = skd(circuit)

        included_gates = set(discretized.count_ops().keys())
        self.assertEqual(set(basis_gates), included_gates)

    def test_load_from_file(self):
        """Test loading basic approximations from a file works.

        Regression test of Qiskit/qiskit#12576.
        """
        filename = "approximations.npy"

        with tempfile.TemporaryDirectory() as tmp_dir:
            fullpath = os.path.join(tmp_dir, filename)

            # dump approximations to file
            generate_basic_approximations(basis_gates=["h", "s", "sdg"], depth=3, filename=fullpath)

            # circuit to decompose and reference decomp
            circuit = QuantumCircuit(1)
            circuit.rx(0.8, 0)

            reference = QuantumCircuit(1, global_phase=3 * np.pi / 4)
            reference.h(0)
            reference.s(0)
            reference.h(0)

            # load the decomp and compare to reference
            skd = SolovayKitaev(basic_approximations=fullpath)
            # skd = SolovayKitaev(basic_approximations=filename)
            discretized = skd(circuit)

        self.assertEqual(discretized, reference)


@ddt
class TestGateSequence(QiskitTestCase):
    """Test the ``GateSequence`` class."""

    def test_append(self):
        """Test append."""
        seq = GateSequence([IGate()])
        seq.append(HGate())

        ref = GateSequence([IGate(), HGate()])
        self.assertEqual(seq, ref)

    def test_eq(self):
        """Test equality."""
        base = GateSequence([HGate(), HGate()])
        seq1 = GateSequence([HGate(), HGate()])
        seq2 = GateSequence([IGate()])
        seq3 = GateSequence([HGate(), HGate()])
        seq3.global_phase = 0.12
        seq4 = GateSequence([IGate(), HGate()])

        with self.subTest("equal"):
            self.assertEqual(base, seq1)

        with self.subTest("same product, but different repr (-> false)"):
            self.assertNotEqual(base, seq2)

        with self.subTest("differing global phase (-> false)"):
            self.assertNotEqual(base, seq3)

        with self.subTest("same num gates, but different gates (-> false)"):
            self.assertNotEqual(base, seq4)

    def test_to_circuit(self):
        """Test converting a gate sequence to a circuit."""
        seq = GateSequence([HGate(), HGate(), TGate(), SGate(), SdgGate()])
        ref = QuantumCircuit(1)
        ref.h(0)
        ref.h(0)
        ref.t(0)
        ref.s(0)
        ref.sdg(0)
        # a GateSequence is SU(2), so add the right phase
        z = 1 / np.sqrt(np.linalg.det(Operator(ref)))
        ref.global_phase = np.arctan2(np.imag(z), np.real(z))

        self.assertEqual(seq.to_circuit(), ref)

    def test_adjoint(self):
        """Test adjoint."""
        seq = GateSequence([TGate(), SGate(), HGate(), IGate()])
        inv = GateSequence([IGate(), HGate(), SdgGate(), TdgGate()])

        self.assertEqual(seq.adjoint(), inv)

    def test_copy(self):
        """Test copy."""
        seq = GateSequence([IGate()])
        copied = seq.copy()
        seq.gates.append(HGate())

        self.assertEqual(len(seq.gates), 2)
        self.assertEqual(len(copied.gates), 1)

    @data(0, 1, 10)
    def test_len(self, n):
        """Test __len__."""
        seq = GateSequence([IGate()] * n)
        self.assertEqual(len(seq), n)

    def test_getitem(self):
        """Test __getitem__."""
        seq = GateSequence([IGate(), HGate(), IGate()])
        self.assertEqual(seq[0], IGate())
        self.assertEqual(seq[1], HGate())
        self.assertEqual(seq[2], IGate())

        self.assertEqual(seq[-2], HGate())

    def test_from_su2_matrix(self):
        """Test from_matrix with an SU2 matrix."""
        matrix = np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)
        matrix /= np.sqrt(np.linalg.det(matrix))
        seq = GateSequence.from_matrix(matrix)

        ref = GateSequence([HGate()])

        self.assertEqual(seq.gates, [])
        self.assertTrue(np.allclose(seq.product, ref.product))
        self.assertEqual(seq.global_phase, 0)

    def test_from_so3_matrix(self):
        """Test from_matrix with an SO3 matrix."""
        matrix = np.array([[0, 0, -1], [0, -1, 0], [-1, 0, 0]])
        seq = GateSequence.from_matrix(matrix)

        ref = GateSequence([HGate()])

        self.assertEqual(seq.gates, [])
        self.assertTrue(np.allclose(seq.product, ref.product))
        self.assertEqual(seq.global_phase, 0)

    def test_from_invalid_matrix(self):
        """Test from_matrix with invalid matrices."""
        with self.subTest("2x2 but not SU2"):
            matrix = np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)
            with self.assertRaises(ValueError):
                _ = GateSequence.from_matrix(matrix)

        with self.subTest("not 2x2 or 3x3"):
            with self.assertRaises(ValueError):
                _ = GateSequence.from_matrix(np.array([[1]]))

    def test_dot(self):
        """Test dot."""
        seq1 = GateSequence([HGate()])
        seq2 = GateSequence([TGate(), SGate()])
        composed = seq1.dot(seq2)

        ref = GateSequence([TGate(), SGate(), HGate()])

        # check the product matches
        self.assertTrue(np.allclose(ref.product, composed.product))

        # check the circuit & phases are equivalent
        self.assertTrue(Operator(ref.to_circuit()).equiv(composed.to_circuit()))


@ddt
class TestSolovayKitaevUtils(QiskitTestCase):
    """Test the public functions in the Solovay Kitaev utils."""

    @data(
        _generate_x_rotation(0.1),
        _generate_y_rotation(0.2),
        _generate_z_rotation(0.3),
        np.dot(_generate_z_rotation(0.5), _generate_y_rotation(0.4)),
        np.dot(_generate_y_rotation(0.5), _generate_x_rotation(0.4)),
    )
    def test_commutator_decompose_return_type(self, u_so3: np.ndarray):
        """Test that ``commutator_decompose`` returns two SO(3) gate sequences."""
        v, w = commutator_decompose(u_so3)
        self.assertTrue(is_so3_matrix(v.product))
        self.assertTrue(is_so3_matrix(w.product))
        self.assertIsInstance(v, GateSequence)
        self.assertIsInstance(w, GateSequence)

    @data(
        _generate_x_rotation(0.1),
        _generate_y_rotation(0.2),
        _generate_z_rotation(0.3),
        np.dot(_generate_z_rotation(0.5), _generate_y_rotation(0.4)),
        np.dot(_generate_y_rotation(0.5), _generate_x_rotation(0.4)),
    )
    def test_commutator_decompose_decomposes_correctly(self, u_so3):
        """Test that ``commutator_decompose`` exactly decomposes the input."""
        v, w = commutator_decompose(u_so3)
        v_so3 = v.product
        w_so3 = w.product
        actual_commutator = np.dot(v_so3, np.dot(w_so3, np.dot(np.conj(v_so3).T, np.conj(w_so3).T)))
        self.assertTrue(np.allclose(actual_commutator, u_so3))

    def test_generate_basis_approximation_gates(self):
        """Test the basis approximation generation works for all supported gates.

        Regression test of Qiskit/qiskit-terra#9585.
        """
        basis = ["i", "x", "y", "z", "h", "t", "tdg", "s", "sdg"]
        approx = generate_basic_approximations(basis, depth=2)

        # This mainly checks that there are no errors in the generation (like
        # in computing the inverse as described in #9585), so a simple check is enough.
        self.assertGreater(len(approx), len(basis))


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
================================================================================
BasicProvider: Python-based Simulators (:mod:`qiskit.providers.basic_provider`)
================================================================================

.. currentmodule:: qiskit.providers.basic_provider

A module of Python-based quantum simulators. Simulators can be accessed
via the `BasicProvider` provider, e.g.:

.. code-block::

   from qiskit.providers.basic_provider import BasicProvider

   backend = BasicProvider().get_backend('basic_simulator')


Classes
=======

.. autosummary::
   :toctree: ../stubs/

   BasicSimulator
   BasicProvider
   BasicProviderJob
   BasicProviderError
"""

from .basic_provider import BasicProvider
from .basic_provider_job import BasicProviderJob
from .basic_simulator import BasicSimulator
from .exceptions import BasicProviderError



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Estimator."""

import unittest
from test import QiskitTestCase

import numpy as np

from qiskit.circuit import Parameter, QuantumCircuit
from qiskit.circuit.library import RealAmplitudes
from qiskit.primitives import StatevectorEstimator
from qiskit.primitives.containers.bindings_array import BindingsArray
from qiskit.primitives.containers.estimator_pub import EstimatorPub
from qiskit.primitives.containers.observables_array import ObservablesArray
from qiskit.quantum_info import SparsePauliOp


class TestStatevectorEstimator(QiskitTestCase):
    """Test Estimator"""

    def setUp(self):
        super().setUp()
        self.ansatz = RealAmplitudes(num_qubits=2, reps=2)
        self.observable = SparsePauliOp.from_list(
            [
                ("II", -1.052373245772859),
                ("IZ", 0.39793742484318045),
                ("ZI", -0.39793742484318045),
                ("ZZ", -0.01128010425623538),
                ("XX", 0.18093119978423156),
            ]
        )
        self.expvals = -1.0284380963435145, -1.284366511861733

        self.psi = (RealAmplitudes(num_qubits=2, reps=2), RealAmplitudes(num_qubits=2, reps=3))
        self.params = tuple(psi.parameters for psi in self.psi)
        self.hamiltonian = (
            SparsePauliOp.from_list([("II", 1), ("IZ", 2), ("XI", 3)]),
            SparsePauliOp.from_list([("IZ", 1)]),
            SparsePauliOp.from_list([("ZI", 1), ("ZZ", 1)]),
        )
        self.theta = (
            [0, 1, 1, 2, 3, 5],
            [0, 1, 1, 2, 3, 5, 8, 13],
            [1, 2, 3, 4, 5, 6],
        )

    def test_estimator_run(self):
        """Test Estimator.run()"""
        psi1, psi2 = self.psi
        hamiltonian1, hamiltonian2, hamiltonian3 = self.hamiltonian
        theta1, theta2, theta3 = self.theta
        estimator = StatevectorEstimator()

        # Specify the circuit and observable by indices.
        # calculate [ <psi1(theta1)|H1|psi1(theta1)> ]
        job = estimator.run([(psi1, hamiltonian1, [theta1])])
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.5555572817900956])

        # Objects can be passed instead of indices.
        # Note that passing objects has an overhead
        # since the corresponding indices need to be searched.
        # User can append a circuit and observable.
        # calculate [ <psi2(theta2)|H1|psi2(theta2)> ]
        result2 = estimator.run([(psi2, hamiltonian1, theta2)]).result()
        np.testing.assert_allclose(result2[0].data.evs, [2.97797666])

        # calculate [ <psi1(theta1)|H2|psi1(theta1)>, <psi1(theta1)|H3|psi1(theta1)> ]
        result3 = estimator.run([(psi1, [hamiltonian2, hamiltonian3], theta1)]).result()
        np.testing.assert_allclose(result3[0].data.evs, [-0.551653, 0.07535239])

        # calculate [ [<psi1(theta1)|H1|psi1(theta1)>,
        #              <psi1(theta3)|H3|psi1(theta3)>],
        #             [<psi2(theta2)|H2|psi2(theta2)>] ]
        result4 = estimator.run(
            [(psi1, [hamiltonian1, hamiltonian3], [theta1, theta3]), (psi2, hamiltonian2, theta2)]
        ).result()
        np.testing.assert_allclose(result4[0].data.evs, [1.55555728, -1.08766318])
        np.testing.assert_allclose(result4[1].data.evs, [0.17849238])

    def test_estimator_with_pub(self):
        """Test estimator with explicit EstimatorPubs."""
        psi1, psi2 = self.psi
        hamiltonian1, hamiltonian2, hamiltonian3 = self.hamiltonian
        theta1, theta2, theta3 = self.theta

        obs1 = ObservablesArray.coerce([hamiltonian1, hamiltonian3])
        bind1 = BindingsArray.coerce({tuple(psi1.parameters): [theta1, theta3]})
        pub1 = EstimatorPub(psi1, obs1, bind1)
        obs2 = ObservablesArray.coerce(hamiltonian2)
        bind2 = BindingsArray.coerce({tuple(psi2.parameters): theta2})
        pub2 = EstimatorPub(psi2, obs2, bind2)

        estimator = StatevectorEstimator()
        result4 = estimator.run([pub1, pub2]).result()
        np.testing.assert_allclose(result4[0].data.evs, [1.55555728, -1.08766318])
        np.testing.assert_allclose(result4[1].data.evs, [0.17849238])

    def test_estimator_run_no_params(self):
        """test for estimator without parameters"""
        circuit = self.ansatz.assign_parameters([0, 1, 1, 2, 3, 5])
        est = StatevectorEstimator()
        result = est.run([(circuit, self.observable)]).result()
        np.testing.assert_allclose(result[0].data.evs, [-1.284366511861733])

    def test_run_single_circuit_observable(self):
        """Test for single circuit and single observable case."""
        est = StatevectorEstimator()

        with self.subTest("No parameter"):
            qc = QuantumCircuit(1)
            qc.x(0)
            op = SparsePauliOp("Z")
            param_vals = [None, [], [[]], np.array([]), np.array([[]]), [np.array([])]]
            target = [-1]
            for val in param_vals:
                self.subTest(f"{val}")
                result = est.run([(qc, op, val)]).result()
                np.testing.assert_allclose(result[0].data.evs, target)
                self.assertEqual(result[0].metadata["target_precision"], 0)

        with self.subTest("One parameter"):
            param = Parameter("x")
            qc = QuantumCircuit(1)
            qc.ry(param, 0)
            op = SparsePauliOp("Z")
            param_vals = [
                [np.pi],
                np.array([np.pi]),
            ]
            target = [-1]
            for val in param_vals:
                self.subTest(f"{val}")
                result = est.run([(qc, op, val)]).result()
                np.testing.assert_allclose(result[0].data.evs, target)
                self.assertEqual(result[0].metadata["target_precision"], 0)

        with self.subTest("More than one parameter"):
            qc = self.psi[0]
            op = self.hamiltonian[0]
            param_vals = [
                self.theta[0],
                [self.theta[0]],
                np.array(self.theta[0]),
                np.array([self.theta[0]]),
                [np.array(self.theta[0])],
            ]
            target = [1.5555572817900956]
            for val in param_vals:
                self.subTest(f"{val}")
                result = est.run([(qc, op, val)]).result()
                np.testing.assert_allclose(result[0].data.evs, target)
                self.assertEqual(result[0].metadata["target_precision"], 0)

    def test_run_1qubit(self):
        """Test for 1-qubit cases"""
        qc = QuantumCircuit(1)
        qc2 = QuantumCircuit(1)
        qc2.x(0)

        op = SparsePauliOp.from_list([("I", 1)])
        op2 = SparsePauliOp.from_list([("Z", 1)])

        est = StatevectorEstimator()
        result = est.run([(qc, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [-1])

    def test_run_2qubits(self):
        """Test for 2-qubit cases (to check endian)"""
        qc = QuantumCircuit(2)
        qc2 = QuantumCircuit(2)
        qc2.x(0)

        op = SparsePauliOp.from_list([("II", 1)])
        op2 = SparsePauliOp.from_list([("ZI", 1)])
        op3 = SparsePauliOp.from_list([("IZ", 1)])

        est = StatevectorEstimator()
        result = est.run([(qc, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op2)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc, op3)]).result()
        np.testing.assert_allclose(result[0].data.evs, [1])

        result = est.run([(qc2, op3)]).result()
        np.testing.assert_allclose(result[0].data.evs, [-1])

    def test_run_errors(self):
        """Test for errors"""
        qc = QuantumCircuit(1)
        qc2 = QuantumCircuit(2)

        op = SparsePauliOp.from_list([("I", 1)])
        op2 = SparsePauliOp.from_list([("II", 1)])

        est = StatevectorEstimator()
        with self.assertRaises(ValueError):
            est.run([(qc, op2)]).result()
        with self.assertRaises(ValueError):
            est.run([(qc, op, [[1e4]])]).result()
        with self.assertRaises(ValueError):
            est.run([(qc2, op2, [[1, 2]])]).result()
        with self.assertRaises(ValueError):
            est.run([(qc, [op, op2], [[1]])]).result()
        with self.assertRaises(ValueError):
            est.run([(qc, op)], precision=-1).result()
        with self.assertRaises(ValueError):
            est.run([(qc, 1j * op)], precision=0.1).result()
        with self.subTest("missing []"):
            with self.assertRaisesRegex(ValueError, "An invalid Estimator pub-like was given"):
                _ = est.run((qc, op)).result()

    def test_run_numpy_params(self):
        """Test for numpy array as parameter values"""
        qc = RealAmplitudes(num_qubits=2, reps=2)
        op = SparsePauliOp.from_list([("IZ", 1), ("XI", 2), ("ZY", -1)])
        k = 5
        rng = np.random.default_rng(12)
        params_array = rng.random((k, qc.num_parameters))
        params_list = params_array.tolist()
        params_list_array = list(params_array)
        estimator = StatevectorEstimator()
        target = estimator.run([(qc, op, params_list)]).result()

        with self.subTest("ndarrary"):
            result = estimator.run([(qc, op, params_array)]).result()
            self.assertEqual(len(result[0].data.evs), k)
            np.testing.assert_allclose(result[0].data.evs, target[0].data.evs)

        with self.subTest("list of ndarray"):
            result = estimator.run([(qc, op, params_list_array)]).result()
            self.assertEqual(len(result[0].data.evs), k)
            np.testing.assert_allclose(result[0].data.evs, target[0].data.evs)

    def test_precision_seed(self):
        """Test for precision and seed"""
        estimator = StatevectorEstimator(default_precision=1.0, seed=1)
        psi1 = self.psi[0]
        hamiltonian1 = self.hamiltonian[0]
        theta1 = self.theta[0]
        job = estimator.run([(psi1, hamiltonian1, [theta1])])
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.901141473854881])
        # The result of the second run is the same
        job = estimator.run([(psi1, hamiltonian1, [theta1]), (psi1, hamiltonian1, [theta1])])
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.901141473854881])
        np.testing.assert_allclose(result[1].data.evs, [1.901141473854881])
        # precision=0 implies the exact expectation value
        job = estimator.run([(psi1, hamiltonian1, [theta1])], precision=0)
        result = job.result()
        np.testing.assert_allclose(result[0].data.evs, [1.5555572817900956])

    def test_iter_pub(self):
        """test for an iterable of pubs"""
        estimator = StatevectorEstimator()
        circuit = self.ansatz.assign_parameters([0, 1, 1, 2, 3, 5])
        observable = self.observable.apply_layout(circuit.layout)
        result = estimator.run(iter([(circuit, observable), (circuit, observable)])).result()
        np.testing.assert_allclose(result[0].data.evs, [-1.284366511861733])
        np.testing.assert_allclose(result[1].data.evs, [-1.284366511861733])

    def test_metadata(self):
        """Test for metadata"""
        qc = QuantumCircuit(2)
        qc2 = QuantumCircuit(2)
        qc2.metadata = {"a": 1}
        estimator = StatevectorEstimator()
        result = estimator.run([(qc, "ZZ"), (qc2, "ZZ")], precision=0.1).result()

        self.assertEqual(len(result), 2)
        self.assertEqual(result.metadata, {"version": 2})
        self.assertEqual(
            result[0].metadata, {"target_precision": 0.1, "circuit_metadata": qc.metadata}
        )
        self.assertEqual(
            result[1].metadata, {"target_precision": 0.1, "circuit_metadata": qc2.metadata}
        )

    def test_reset(self):
        """Test for circuits with reset."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)
        qc.reset(0)
        op = SparsePauliOp("ZI")

        seed = 12
        n = 1000
        estimator = StatevectorEstimator(seed=seed)
        with self.subTest("precision=0"):
            result = estimator.run([(qc, [op] * n)]).result()
            # expectation values should be stochastic due to reset for subsystems
            np.testing.assert_allclose(result[0].data.evs.mean(), 0, atol=1e-1)

            result2 = estimator.run([(qc, [op] * n)]).result()
            # expectation values should be reproducible due to seed
            np.testing.assert_allclose(result[0].data.evs, result2[0].data.evs)

        with self.subTest("precision=0.01"):
            precision = 0.01
            result = estimator.run([(qc, [op] * n)], precision=precision).result()
            # expectation values should be stochastic due to reset for subsystems
            np.testing.assert_allclose(result[0].data.evs.mean(), 0, atol=1e-1)

            result2 = estimator.run([(qc, [op] * n)], precision=precision).result()
            # expectation values should be reproducible due to seed
            np.testing.assert_allclose(result[0].data.evs, result2[0].data.evs)


if __name__ == "__main__":
    unittest.main()
